#
#  opt_func.tcl -- Generic optimization procedure functions.
#
#  Written by William Lavender.
#
#---------------------------------------------------------------------
#
#  Copyright 1999 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#


proc setup_optimize_database { motor_database scan_database } {

	#
	# Create a new record list.
	#

	set record_list [ Mx::RecordList ::mxrecordlist_#auto ]

	#
	# Enable plotting.
	#

	$record_list set_plot_enable 1

	#
	# Read the motor and scan databases.
	#

	$record_list read_database $motor_database
	$record_list read_database $scan_database

	#
	# Finish initializing MX.
	#

	$record_list finish_database_initialization
	$record_list initialize_hardware

	#
	# Install an event handler to allow Tk events to be processed
	# while MX is scanning.
	#

	Mx::install_user_interrupt_handler

	#
	# Install handler for mx_info_dialog() for things like
	# end of scan dialogs.
	#

	Mx::install_dialog_handler

	return $record_list
}

proc perform_optimization { record_list motor guifield } {

	set motor_name [ $motor get_field "name" ]

	#
	# For each motor there should be a scan with the same name, but
	# with _scan tacked onto the end of it.
	#

	set scan_name ${motor_name}_scan

	set status [ catch { $record_list get_record $scan_name } scan_record ]

	if { $status != 0 } {
		set t "The scan '$scan_name' does not exist.  "
		append t "I cannot perform the optimization without it."

		tk_dialog .scan_does_not_exist_dialog "Scan Record Missing" \
			$t error 0 "OK"

		return
	}

	#
	# Save the old motor position.
	#

		set old_position [ $motor get_position ]

		set units [ $motor get_field "units" ]

	#
	# Update the screen display with the most current old position.
	#

		${guifield}_old_position configure -text $old_position

	#
	# Calculate the scan range.
	#
		set step_size [ $scan_record get_field "step_size" ]
		set num [ $scan_record get_field "num_measurements" ]

		set scan_range [ expr ($num - 1) * $step_size ]

	#
	# Change the scan start position so that the scan range is centered
	# on the original position of the motor.
	#

		set scan_start [ expr $old_position - 0.5 * $scan_range ]

		$scan_record set_field "start_position" $scan_start
	#
	# Run the scan.
	#

	${guifield}_status configure -fg yellow \
			-text "Optimization scan in progress"

	set status [ catch { $scan_record perform_scan } ]

	if { $status != 0 } {
		${guifield}_status configure -fg hotpink \
			-text "Moving to original start position."

		log_puts \
"Optimization scan failed.  Moving to old position $old_position $units"

		$motor move_absolute $old_position

		after 1000 \
			check_final_move_status $motor $guifield "scan_failed"

		set t "The scan '$scan_name' failed with status = $status.  "
		append t "Moving '$motor_name' back to its original position."

		tk_dialog .scan_failed_dialog "Scan Failed" \
			$t error 0 "OK"

		return
	}

	#
	# Perform the fit with an external script.
	#

	set program fit_to_max

	set datafile [ $scan_record get_field "datafile" ]

	set datafile_list [ split $datafile ]

	set datafile_name [ lindex $datafile_list 1 ]

	#

	set command "$program $datafile_name"

	log_puts "fitting command = '$command'"

	${guifield}_status configure -fg yellow -text "Fitting in progress"

	#
	# Run the script.
	#

	set status [ catch { eval exec $command } result ]

	log_puts "status = '$status', result = '$result'"

	if { $status != 0 } {
		${guifield}_status configure -fg hotpink \
				-text "Moving to original start position"

		log_puts \
"Fitting program crashed.  Moving to old position $old_position $units"

		$motor move_absolute $old_position

		after 1000 \
			check_final_move_status $motor $guifield "fit_crashed"

		set t "The external fit program crashed with status = $status."
		append t "  Moving '$motor_name' back to its "
		append t "position."

		tk_dialog .fit_crashed_dialog "Fit Crashed" \
			$t error 0 "OK"

		return
	}

	#
	# Did the fit work?  If so, move the motor to the new position.
	#

	set result_list [ split $result ]

	set fit_status   [ lindex $result_list 0 ]
	set new_position [ lindex $result_list 1 ]

	if { $fit_status == 0 } {
		${guifield}_status configure -fg hotpink \
				-text "Moving to original start position"

		log_puts \
"Optimization fit failed.  Moving to old position $old_position $units"

		$motor move_absolute $old_position

		after 1000 \
			check_final_move_status $motor $guifield "fit_failed"
	} else {
		set dialog .optimization_succeeded_dialog

		set dialog_font [ font create -family Times -size 15 ]

		toplevel $dialog
		wm title $dialog "Optimization Succeeded"

		label $dialog.message1 -font $dialog_font \
				-text "The optimization was successful."
		pack $dialog.message1

		set mtext "The new suggested position for '$motor_name' "
		append mtext "is $new_position $units."

		label $dialog.message2 -font $dialog_font -text $mtext
		pack $dialog.message2

		frame $dialog.sep -height 2 -borderwidth 1 -relief sunken
		pack $dialog.sep -expand yes -fill both

		iwidgets::buttonbox $dialog.bb
		pack $dialog.bb -expand yes -fill both

		$dialog.bb add Accept -text Accept -command [ code \
			optimization_succeeded_accept $dialog $guifield \
			$motor $new_position $units ]

		$dialog.bb add Discard -text Discard -command [ code \
			optimization_succeeded_discard $dialog $guifield \
			$motor $old_position $units ]

		$dialog.bb default Accept
	}
}

proc optimization_succeeded_accept { dialog guifield \
				motor new_position units } {

	destroy $dialog

	${guifield}_status configure -fg springgreen \
				-text "Moving to new position."

	${guifield}_new_position configure -text $new_position

	log_puts \
"Optimization succeeded.  Moving to optimized position $new_position $units"

	$motor move_absolute $new_position

	after 1000 check_final_move_status $motor $guifield "success"
}

proc optimization_succeeded_discard { dialog guifield \
				motor old_position units } {

	destroy $dialog

	${guifield}_status configure -fg purple \
				-text "Moving to original start position."

	${guifield}_new_position configure -text $old_position

	set message "Optimization succeeded, but was discarded.  "
	append message "Moving to original position $old_position $units"

	log_puts $message

	$motor move_absolute $old_position

	after 1000 check_final_move_status $motor $guifield "discarded"
}

proc check_final_move_status { motor guifield condition } {

	if [ $motor is_busy ] {

		after 1000 check_final_move_status $motor $guifield $condition

		return
	}

	if { $condition == "scan_failed" } {

		${guifield}_status configure -fg red \
				-text "Optimization scan failed."

	} elseif { $condition == "fit_crashed" } {

		${guifield}_status configure -fg red \
				-text "Fitting program crashed."

	} elseif { $condition == "fit_failed" } {

		${guifield}_status configure -fg red \
				-text "Optimization fit failed."

	} elseif { $condition == "success" } {

		${guifield}_status configure -fg limegreen \
				-text "Optimization succeeded."

	} elseif { $condition == "discarded" } {

		${guifield}_status configure -fg maroon \
				-text "Optimization discarded."

	} else { 
		${guifield}_status configure -fg red \
	-text "check_final_move_status invoked with condition '$condition'"
	}
}

proc create_message_log_window {} {

	set window .message_log

        #
        #  Create the window.
        #

        iwidgets::scrolledtext $window -labeltext "Message Log"

        $window configure -visibleitems 40x10
        $window configure -hscrollmode dynamic
        $window configure -vscrollmode dynamic

        pack $window -expand yes -fill both

        #
        #  Install handlers to redirect the output of mx_info(), mx_error()
        #  and MX_DEBUG() to this window.
        #

        Mx::install_output_handlers $window

	log_puts "*** Optimize GUI Initializing ***"
}

proc log_puts { message } {
	set textwindow [ .message_log childsite ]

	$textwindow configure -state normal

	.message_log insert end "$message\n"

	$textwindow configure -state disabled

	.message_log yview moveto 1.0
}

